import  java.util.Collections;
import  java.util.HashMap;
import  java.util.Map;
import  java.util.Properties;
import  java.util.Random;

import  javax.jms.Connection;
import  javax.jms.ConnectionFactory;
import  javax.jms.DeliveryMode;
import  javax.jms.Destination;
import  javax.jms.JMSException;
import  javax.jms.MapMessage;
import  javax.jms.MessageConsumer;
import  javax.jms.MessageProducer;
import  javax.jms.Session;

import  javax.naming.Context;
import  javax.naming.InitialContext;
import  javax.naming.NamingException;

import  org.apache.log4j.BasicConfigurator;
import  org.apache.log4j.ConsoleAppender;
import  org.apache.log4j.Level;
import  org.apache.log4j.Logger;
import  org.apache.log4j.PatternLayout;

/**
 * JMS load test class
 * @author rschuetz
 */

public class LoadTest
{
 // === Static ================================================================

 private static final Logger         LOGGER          = Logger.getLogger(LoadTest.class);
 private static final long           RECEIVE_TIMEOUT = 30000L;

 private static final String         ID_MESSAGEID    = "messageid",
                                     ID_SEQUENCEID   = "sequenceid",
                                     ID_TIMESTAMP    = "timestamp",
                                     ID_PAYLOAD      = "payload",
                                     ID_PAYLOADID    = "payloadid";
 
 // === Attributes ============================================================

 private final Map<Integer, byte[]>  MESSAGES;

 private final Random                RANDOM;

 private final String                JNDIFACTORYINITIAL,
                                     JNDIPROVIDERURL,
                                     JNDICONNECTIONFACTORYALIAS,
                                     JNDIQUEUEALIAS;

 private final long                  STARTTIME,
                                     TOTALRUNTIME;

 private final int                   MSGCOUNT,
                                     MSGCOUNTVARIANCE,
                                     STAGELOOPS,
                                     STAGES,
                                     MINMSGSIZE,
                                     MAXMSGSIZE,
                                     COMMITINTERVAL,
                                     RECONNECTINTERVAL;
 
 private final boolean               PERSISTANT_STORAGE;

 private       Connection            producerConnection       = null,
                                     consumerConnection       = null;
 private       Session               producerSession          = null,
                                     consumerSession          = null;
 private       MessageProducer       messageProducer          = null;
 private       MessageConsumer       messageConsumer          = null;

 private       long                  totalNbrMessagesSent     = 0L,
                                     totalNbrMessagesReceived = 0L,
                                     msgSequenceID            = 0L,
                                     queueLength              = 0L,
                                     totalNbrMessagesLost     = 0L,
                                     totalNbrMessagesDupl     = 0L;

 // === Constructor ===========================================================

 public LoadTest(String[] args)
 {
  int                   argIdx;
  long                  seed;

  byte[]                message;

  Map <Integer, byte[]> messages;

  // Initialize logger

  BasicConfigurator.resetConfiguration();

  Logger.getRootLogger().addAppender(new ConsoleAppender(new PatternLayout("%-5p [%d{ISO8601}] %m%n")));
  Logger.getRootLogger().setLevel(Level.INFO);

  // Get parameters

  if(args.length == 15)
  {
   argIdx                     = 0;

   JNDIFACTORYINITIAL         = args[argIdx ++];
   JNDIPROVIDERURL            = args[argIdx ++];
   JNDICONNECTIONFACTORYALIAS = args[argIdx ++];
   JNDIQUEUEALIAS             = args[argIdx ++];

   seed                       = Long   .parseLong   (args[argIdx ++]);
   TOTALRUNTIME               = Long   .parseLong   (args[argIdx ++]);
   MSGCOUNT                   = Integer.parseInt    (args[argIdx ++]);
   MSGCOUNTVARIANCE           = Integer.parseInt    (args[argIdx ++]);
   STAGES                     = Integer.parseInt    (args[argIdx ++]);
   STAGELOOPS                 = Integer.parseInt    (args[argIdx ++]);
   MINMSGSIZE                 = Integer.parseInt    (args[argIdx ++]);
   MAXMSGSIZE                 = Integer.parseInt    (args[argIdx ++]);
   COMMITINTERVAL             = Integer.parseInt    (args[argIdx ++]);
   RECONNECTINTERVAL          = Integer.parseInt    (args[argIdx ++]);
   PERSISTANT_STORAGE         = Boolean.parseBoolean(args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + LoadTest.class.getSimpleName() + 
                                      " <jndi factory initial> <jndi provider URL> <jndi connection factory alias> <jndi queue alias>" +
   		                              " <rng seed> <total runtime in ms> <max. message count in a queue (+/- variance)> <message count variance>" +
   		                              " <# of stages> <# of loops adding/removing # of variance messages on queue per stage>" +
   		                              " <minimum message size in bytes> <maximum message size in bytes>" +
   		                              " <commit interval in # of messages processed> <reconnect interval in # of messages processed>" +
   		                              " <persistant storage (true/false)>");
  }

  // Initialize
  
  STARTTIME     = System.currentTimeMillis();
  queueLength   = 0L;
  msgSequenceID = 0L;
  
  // Initialize RNG. Fixed seed for maybe deterministic reproducibility

  seed = (seed == 0L ? System.nanoTime() + 8682522807148012L : seed);
  RANDOM = new Random(seed);

  LOGGER.info("Seed: " + seed);

  // Pre-create messages

  messages = new HashMap<Integer, byte[]>();

  for(int i = MINMSGSIZE; i <= MAXMSGSIZE; i ++)
  {
   message = new byte[i];
   RANDOM.nextBytes(message);
   messages.put(i, message);
  }

  MESSAGES = Collections.unmodifiableMap(messages);
 }

 // === Public ================================================================

 public static void main(String[] args)
 throws Exception
 {
  (new LoadTest(args)).runLoadTest();
 }
 
 public void runLoadTest()
 {
  int  mainloop, stage;
  long stageLimit, nbrMessagesToSend, nbrMessagesToReceive;

  try
  {
   // Initial connect
 
   connect();
 
   // Loop until specified runtime elapsed
 
   for(mainloop = 0; System.currentTimeMillis() < STARTTIME + TOTALRUNTIME; mainloop ++)
   {
    // Fill queue in stages
 
    for(stage = 0; stage <= STAGES; stage ++)
    {
     // Fill queue to stage limit
 
     stageLimit        = (MSGCOUNT * stage) / STAGES;
     nbrMessagesToSend = stageLimit - queueLength;

     LOGGER.info("FILL: Filling queue: Loop: " + mainloop + " / " + stage + "; limit " + stageLimit);

     if(nbrMessagesToSend > 0) 
     {
      sendMessages(nbrMessagesToSend);
     }
     else if(nbrMessagesToSend < 0)
     {
      // Queue is longer than expected; most likely caused by duplicate messages
      
      LOGGER.info("FILL: Correcting queue length; receiving " + (-nbrMessagesToSend) + " messages additionally.");
      receiveMessages(-nbrMessagesToSend);
     }
     
     // Apply variance
 
     for(int stageLoop = 0; stageLoop < STAGELOOPS; stageLoop ++)
     {
      if(stageLimit - MSGCOUNTVARIANCE < 0)
      {
       LOGGER.info("FILL/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; sending.");
       sendMessages(MSGCOUNTVARIANCE);
 
       LOGGER.info("FILL/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; receiving.");
       receiveMessages(MSGCOUNTVARIANCE);
      }
      else
      {
       LOGGER.info("FILL/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; sending [1].");
       sendMessages(MSGCOUNTVARIANCE);
 
       LOGGER.info("FILL/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; receiving.");
       receiveMessages(MSGCOUNTVARIANCE * 2);
 
       LOGGER.info("FILL/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; sending [2].");
       sendMessages(MSGCOUNTVARIANCE);
      }
     }
    }
 
    // Print totals
 
    LOGGER.info("FILL: Totals: # msgs sent: " + totalNbrMessagesSent + " # msgs recv: " + totalNbrMessagesReceived);
 
    // Empty queue in stages
 
    for(stage = STAGES; stage >= 0; stage --)
    {
     // Empty queue to stage limit
 
     stageLimit           = (MSGCOUNT * stage) / STAGES;
     nbrMessagesToReceive = queueLength - stageLimit;
 
     LOGGER.info("EMPY: Emptying queue: Loop: " + mainloop + " / " + stage + "; limit " + stageLimit);
 
     if(nbrMessagesToReceive > 0) 
     {
      receiveMessages(nbrMessagesToReceive);
     }
     else if(nbrMessagesToReceive < 0)
     {
      // Queue is shorter than expected; most likely caused by lost messages
      
      LOGGER.info("EMPY: Correcting queue length; sending " + (-nbrMessagesToReceive) + " messages additionally.");
      sendMessages(-nbrMessagesToReceive);
     }     
 
     // Apply variance
 
     for(int stageLoop = 0; stageLoop < STAGELOOPS; stageLoop ++)
     {
      if(stageLimit - MSGCOUNTVARIANCE < 0)
      {
       LOGGER.info("EMPY/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; sending.");
       sendMessages(MSGCOUNTVARIANCE);
 
       LOGGER.info("EMPY/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; receiving.");
       receiveMessages(MSGCOUNTVARIANCE);
      }
      else
      {
       LOGGER.info("EMPY/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; sending [1].");
       sendMessages(MSGCOUNTVARIANCE);
 
       LOGGER.info("EMPY/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; receiving.");
       receiveMessages(MSGCOUNTVARIANCE * 2);
 
       LOGGER.info("EMPY/VAR: Loop: " + mainloop + " / " + stage + " / " + stageLoop + "; sending [2].");
       sendMessages(MSGCOUNTVARIANCE);
      }
     }
    }
 
    // Print totals
 
    LOGGER.info("EMPY: Totals: # msgs sent: " + totalNbrMessagesSent + " # msgs recv: " + totalNbrMessagesReceived);
   }
  }
  catch(Exception e)
  {
   LOGGER.error("Exception caught", e);
  }
  finally
  {
   try
   {
    disconnect(); 
   }
   catch(Exception e)
   {
    LOGGER.error("Exception caught while disconnecting", e);
   }
  }
 }

 private void sendMessages(final long nbrMessages) 
 throws InterruptedException
 {
  boolean             proceed, reconnect, commit;
  
  byte[]              payload;
  int                 payloadId;
  
  long                time, nbrMessagesSent, nbrUncommitedMessages;
  MapMessage          mapMessage;

  // Log

  LOGGER.info("Sending " + nbrMessages + " messages, current queue size: " + queueLength);

  // Process messages

  nbrMessagesSent       = 0L;
  nbrUncommitedMessages = 0L;
  time                  = System.currentTimeMillis();  

  while(nbrMessagesSent < nbrMessages)
  {
   // Select payload

   payloadId = RANDOM.nextInt(MESSAGES.size());
   payload   = MESSAGES.get(payloadId);

   // Create & send message

   proceed = false;
   
   do
   {   
    try
    {
     mapMessage = producerSession.createMapMessage();
 
     mapMessage.setStringProperty(ID_MESSAGEID,  this.getClass().getSimpleName() + "-" + STARTTIME + "#" + totalNbrMessagesSent);
     mapMessage.setLongProperty  (ID_TIMESTAMP,  System.currentTimeMillis());
     mapMessage.setLongProperty  (ID_SEQUENCEID, msgSequenceID);
    
     mapMessage.setBytes         (ID_PAYLOAD,    payload);
     mapMessage.setInt           (ID_PAYLOADID,  payloadId);
 
     messageProducer.send(mapMessage);
     
     proceed = true;
    }
    catch(Exception e)
    {
     LOGGER.error("Exception while creating/sending message; retrying.", e);
     Thread.sleep(1000);
    }
   } while(!proceed);
   
   // Adapt counters

   nbrMessagesSent       ++;
   totalNbrMessagesSent  ++;
   
   msgSequenceID         ++;
   nbrUncommitedMessages ++;
   
   queueLength           ++;

   // Commit/Reconnect?
   
   reconnect = (totalNbrMessagesSent + totalNbrMessagesReceived) % RECONNECTINTERVAL == 0;
   commit    = reconnect || totalNbrMessagesSent % COMMITINTERVAL == 0 || nbrMessagesSent == nbrMessages;
   
   // Commit

   if(commit)
   {
    if(COMMITINTERVAL >= 1000) LOGGER.info("Committing producer session; sequenceID = " + msgSequenceID + " queueLength = " + queueLength + " nbrUncommitedMessages = " + nbrUncommitedMessages);
    
    try
    {
     producerSession.commit();
    }
    catch(JMSException e)
    {
     LOGGER.error("JMSException while committing sent messages; resending messages.", e);
     
     // Commit failed, correct counters, messages must be resend
     
     nbrMessagesSent      -= nbrUncommitedMessages;
     totalNbrMessagesSent -= nbrUncommitedMessages;
     msgSequenceID        -= nbrUncommitedMessages;
     queueLength          -= nbrUncommitedMessages;     

     // Rollback
     
     try
     {
      producerSession.rollback();
     }
     catch(Exception e1)
     {
      throw new IllegalStateException("rollback() failed after commit(); stopping ...");
     }
     
     // Reconnect (test to see, if server becomes responsive again ...)
     
     try
     {
      reconnect();
     }
     catch(Exception e1)
     {
      throw new IllegalStateException("reconnect() failed after rollback(); stopping ...");
     }
    }
    finally
    {
     nbrUncommitedMessages = 0L;
    }
   }
   
   if(reconnect)
   {
    // Disconnect/Reconnect

    proceed = false;
    
    do
    {
     try
     {
      if((totalNbrMessagesSent + totalNbrMessagesReceived) % RECONNECTINTERVAL == 0) reconnect();
      proceed = true;
     }
     catch(Exception e)
     {
      LOGGER.error("Exception while reconnecting; retrying.", e);
      Thread.sleep(1000);
     }
    } while(!proceed);
   }
  }

  time = System.currentTimeMillis() - time;  
  
  // Check

  if(nbrMessagesSent != nbrMessages) throw new IllegalStateException("Send: expected: " + nbrMessages + " actual: " + nbrMessagesSent);

  // Log

  LOGGER.info("Done. " + nbrMessagesSent + " messages sent, current queue size: " + queueLength + " speed: " + ((1000L * nbrMessagesSent) / time) + " msgs/s");
  LOGGER.info("Totals: # msgs sent: " + totalNbrMessagesSent + " # msgs recv: " + totalNbrMessagesReceived + " # msgs lost: " + totalNbrMessagesLost + " # msgs dupl: " + totalNbrMessagesDupl); 
 }

 private void receiveMessages(final long nbrMessages)
 throws InterruptedException
 {
  boolean             proceed, reconnect, commit;
  
  long                nbrMessagesReceived, nbrUncommitedMessages, expMessageSequenceID, messageSequenceID, gap, time;
  String              messageId;

  MapMessage          mapMessage;

  // Log

  LOGGER.info("Receiving " + nbrMessages + " messages, current queue size: " + queueLength);

  // Process messages

  nbrMessagesReceived   = 0L;
  nbrUncommitedMessages = 0L;
  time                  = System.currentTimeMillis();

  while(nbrMessagesReceived < nbrMessages)
  {
   // Receive

   proceed           = false;
   
   mapMessage        = null;
   messageSequenceID = 0L;
   messageId         = null; 
   
   do
   {
    try
    {
     if((mapMessage = ((MapMessage)(messageConsumer.receive(RECEIVE_TIMEOUT)))) != null)
     {
      messageSequenceID = mapMessage.getLongProperty  (ID_SEQUENCEID);
      messageId         = mapMessage.getStringProperty(ID_MESSAGEID);
      proceed           = true;
     }
     else
     {
      // No message after timeout; reconnect might help
      
      LOGGER.error("messageConsumer#receive() did not return any new messages. Reconnecting.");
      
      try
      {
       reconnect();
      }
      catch(Exception e1)
      {
       throw new IllegalStateException("reconnect() failed after receive() did not return any new messages; stopping ...");
      }  
     }
    }
    catch(Exception e)
    {
     LOGGER.error("Exception while receiving message; retrying.", e);
     Thread.sleep(1000);
    }
   } while(!proceed);
   
   // Message received
   
   if(mapMessage != null)
   {
    // Compare sequence IDs
    
    expMessageSequenceID = msgSequenceID - queueLength;
    gap                  = messageSequenceID - expMessageSequenceID;

    if(gap != 0)
    {
     // Message sequence broken. Either we've duplicate messages or we lost messages ...
     
     LOGGER.error("Received message != expected message. recv: " + messageSequenceID + " exp: " + expMessageSequenceID + " gap: " + gap + " recvMsgId: " + messageId + " # msgs recv so far: " + totalNbrMessagesReceived);

     // Correct queue length to continue testing
     // gap > 0: messages lost, gap < 0 duplicate messages
     
     if(gap > 0) { totalNbrMessagesLost +=  gap; queueLength -=  gap; }
            else { totalNbrMessagesDupl += -gap; queueLength += -gap; }
    }
    
    // Update counters

    nbrMessagesReceived      ++;
    totalNbrMessagesReceived ++;
    nbrUncommitedMessages    ++;
    queueLength              --;
   }
   
   // Commit/Reconnect?
   
   reconnect = (totalNbrMessagesSent + totalNbrMessagesReceived) % RECONNECTINTERVAL == 0;
   commit    = reconnect || totalNbrMessagesReceived % COMMITINTERVAL == 0 || nbrMessagesReceived == nbrMessages;   
   
   // Commit

   if(commit)
   {
    // Commit
    
    if(COMMITINTERVAL >= 1000) LOGGER.info("Committing consumer session; sequenceID = " + msgSequenceID + " queueLength = " + queueLength + " nbrUncommitedMessages = " + nbrUncommitedMessages);
    
    try
    {
     consumerSession.commit();
    }
    catch(JMSException e)
    {
     LOGGER.error("JMSException while committing received messages; re-receiving messages.", e);
     
     // Commit failed, correct counters, ALL messages should be re-received.

     nbrMessagesReceived      -= nbrUncommitedMessages;
     totalNbrMessagesReceived -= nbrUncommitedMessages;
     queueLength              += nbrUncommitedMessages;     
     
     // Rollback
      
     try
     {
      producerSession.rollback();
     }
     catch(Exception e1)
     {
      throw new IllegalStateException("rollback() failed after commit(); stopping ...");
     }
    }
    finally
    {
     nbrUncommitedMessages = 0L;
    }
   }
   
   if(reconnect)
   {
    // Disconnect/Reconnect

    proceed = false;
    
    do
    {
     try
     {
      if((totalNbrMessagesSent + totalNbrMessagesReceived) % RECONNECTINTERVAL == 0) reconnect();
      proceed = true;
     }
     catch(Exception e)
     {
      LOGGER.error("Exception while reconnecting; retrying.", e);
      Thread.sleep(1000);
     }
    } while(!proceed);        
   }
  }

  time = System.currentTimeMillis() - time;
  
  // Check

  if(nbrMessagesReceived != nbrMessages) throw new IllegalStateException("Receive: expected: " + nbrMessages + " actual: " + nbrMessagesReceived);

  // Log

  LOGGER.info("Done. " + nbrMessagesReceived + " messages received, current queue size: " + queueLength + " speed: " + ((1000L * nbrMessagesReceived) / time) + " msgs/s");
  LOGGER.info("Totals: # msgs sent: " + totalNbrMessagesSent + " # msgs recv: " + totalNbrMessagesReceived + " # msgs lost: " + totalNbrMessagesLost + " # msgs dupl: " + totalNbrMessagesDupl);
 }

 // === Private ===============================================================

 private void reconnect()
 throws JMSException,
        NamingException
 {
  disconnect();
  connect();
 }

 private void connect()
 throws JMSException,
        NamingException
 {
  Properties        jndiProperties;
  InitialContext    jndiContext;
  ConnectionFactory connectionFactory;
  Destination       destination;

  // Log

  LOGGER.info("Connecting.");

  // Initialize JNDI

  jndiProperties = new Properties();
  jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, JNDIFACTORYINITIAL);
  jndiProperties.setProperty(Context.PROVIDER_URL,            JNDIPROVIDERURL);

  jndiContext = new InitialContext(jndiProperties);

  // Lookup connection factory and queue

  connectionFactory = ((ConnectionFactory)(jndiContext.lookup(JNDICONNECTIONFACTORYALIAS)));
  destination       = ((Destination      )(jndiContext.lookup(JNDIQUEUEALIAS)));

  // Create producer connection, session and message producer

  producerConnection = connectionFactory.createConnection();
  producerSession    = producerConnection.createSession(true, 0);

  messageProducer = producerSession.createProducer(destination);
  messageProducer.setDeliveryMode(PERSISTANT_STORAGE ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);

  producerConnection.start();

  // Create consumer connection, session and message consumer

  consumerConnection = connectionFactory.createConnection();
  consumerSession    = consumerConnection.createSession(true, 0);

  messageConsumer = consumerSession.createConsumer(destination);

  consumerConnection.start();

  // Log

  LOGGER.info("Connected.");
 }

 private void disconnect()
 throws JMSException
 {
  // Log

  LOGGER.info("Disconnecting.");

  // Close assets

  if(messageProducer    != null) { messageProducer   .close(); messageProducer    = null; }
  if(messageConsumer    != null) { messageConsumer   .close(); messageConsumer    = null; }
  if(producerSession    != null) { producerSession   .close(); producerSession    = null; }
  if(consumerSession    != null) { consumerSession   .close(); consumerSession    = null; }
  if(producerConnection != null) { producerConnection.close(); producerConnection = null; }
  if(consumerConnection != null) { consumerConnection.close(); consumerConnection = null; }

  // Log

  LOGGER.info("Disconnected.");
 }

 // === Inner classes =========================================================
}
